#------------------------------------------------------------------
# $Copyright: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/WebSphereMQ6/WebSphereMQ6.pm,v 1.33 2011/05/09 11:34:40 lpatil Exp $
#
# $Author: lpatil $
# $Date: 2011/05/09 11:34:40 $
# $Revision: 1.33 $
#
# Provides:                            WebSphereMQ6
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package WebSphereMQ6;

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VAD home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VAD..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvad";
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /(...)$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use File::Path qw ( mkpath );
use Fcntl;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: WebSphereMQ6.pm,v 1.33 2011/05/09 11:34:40 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

use constant EPOP           => 10;
use constant CHECK_INTERVAL => 5;

#-----------------------------------
# Start command and start arguments
#-----------------------------------
use constant START_COMMAND                => "/opt/mqm/bin/strmqm";
use constant START_COMMAND_CMDSERVER      => "/opt/mqm/bin/strmqcsv";
use constant START_COMMAND_AIX_CMDSERVER  => "/usr/mqm/bin/strmqcsv";
use constant START_AIX_COMMAND            => "/usr/mqm/bin/strmqm";
use constant START_COMMAND_ARGS           => "X_QUEUE_MANAGER_X";

#----------------------------------
# Stop command and stop arguments
#----------------------------------
use constant STOP_COMMAND                   => "/opt/mqm/bin/endmqm";
use constant STOP_AIX_COMMAND               => "/usr/mqm/bin/endmqm";
use constant STOP_IMMEDIATE_COMMAND_ARGS    => "-i X_QUEUE_MANAGER_X";
use constant STOP_PREEMPTIVE_COMMAND_ARGS   => "-p X_QUEUE_MANAGER_X";

#-------------------------------------------
# Clear IPC command (different path on AIX)
#-------------------------------------------
use constant CLEAN_IPC_COMMAND              => "/opt/mqm/bin/amqiclen";
use constant CLEAN_IPC_AIX_COMMAND          => "/usr/mqm/bin/amqiclen";
use constant CLEAN_IPC_COMMAND_ARGS         => "-x -m X_QUEUE_MANAGER_X";

#--------------------------------
# Second level monitor command
#--------------------------------
use constant SECOND_LEVEL_MONITOR_COMMAND       => "/opt/mqm/bin/runmqsc";
use constant SECOND_LEVEL_MONITOR_AIX_COMMAND   => "/usr/mqm/bin/runmqsc";
use constant SECOND_LEVEL_MONITOR_COMMAND_ARGS  => "X_QUEUE_MANAGER_X >/dev/null <<END1
ping qmgr
END1";


#-------------------------------------------------------------------------------
# Define WebSphere MQ processes to monitor
#-------------------------------------------------------------------------------
use constant MONITOR_PROC_LIST_53 => (
                                        "amqhasmx X_QUEUE_MANAGER_X( |\$)",
                                        "amqzllp0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlaa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                      );
use constant MONITOR_PROC_LIST_60 => (
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmuc0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmur0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                      );
use constant COMMAND_SERVER_PROC  => "amqpcsea *X_QUEUE_MANAGER_X( |\$)";
#
# http://publib.boulder.ibm.com/infocenter/wmqv6/v6r0/index.jsp?topic=/com.ibm.mq.amqzag.doc/fa22310_.htm
# End the processes in the following order:
#
# amqzmuc0 	Critical process manager
# amqzxma0 	Execution controller
# amqzfuma 	OAM process
# amqzlaa0 	LQM agents
# amqzlsa0 	LQM agents
# amqzmgr0      Process controller
# amqzmur0 	Restartable process manager
# amqrmppa 	Process pooling process
# amqrrmfa 	The repository process (for clusters)
# amqzdmaa 	Deferred message processor
# amqpcsea 	The command server
#

use constant KILL_PROC_LIST_60 => (
                                        "amqzmuc0 .*-m *X_QUEUE_MANAGER_X( |\$)",  
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlaa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlsa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmgr0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmur0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrmppa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqpcsea *X_QUEUE_MANAGER_X( |\$)",
                                        "amqhasmx X_QUEUE_MANAGER_X( |\$)",
                                        "amqzllp0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqharmx .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqfqpub .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqfcxba .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqcrsta .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqsc *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                   );
use constant KILL_PROC_LIST_53 => (
                                        "amqhasmx X_QUEUE_MANAGER_X( |\$)",
                                        "amqzllp0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlaa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqpcsea .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqharmx .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                   );
use constant MONITOR_PROC_LIST_70 => ( 
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmuc0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmur0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                      ); 
#
# http://publib.boulder.ibm.com/infocenter/wmqv7/v7r0/index.jsp?topic=/com.ibm.mq.amqzag.doc/fa22310_.htm
# End the processes in the following order:
#
# amqzmuc0 	Critical process manager
# amqzxma0 	Execution controller
# amqzfuma 	OAM process
# amqzlaa0 	LQM agents
# amqzlsa0 	LQM agents
# amqzmuf0 	Utility Manager
# amqzmur0 	Restartable process manager
# amqzmgr0      Process controller
# amqfqpub 	Publish Subscribe process
# amqfcxba 	Broker worker process
# amqrmppa 	Process pooling process
# amqcrsta      Non-threaded responder job process
# amqcrs6b      LU62 receiver channel and client connection
# amqrrmfa 	The repository process (for clusters)
# amqzdmaa 	Deferred message processor
# amqpcsea 	The command server
# runmqtrm      Invoke a trigger monitor for a server
# runmqdlq      Invoke dead-letter queue handler
# runmqchi 	The channel initiator process
# runmqlsr 	The channel listener process
# amqxssvn 	Shared memory servers
# amqztrcn 	Trace
#

use constant KILL_PROC_LIST_70 => ( 
                                        "amqzmuc0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzxma0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzfuma .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlaa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzlsa0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmuf0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmur0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzmgr0 .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqfqpub .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqfcxba .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrmppa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqcrsta .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqcrs6b .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqrrmfa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqzdmaa .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqpcsea *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqtrm .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqdlq .*X_QUEUE_MANAGER_X( |\$)",
                                        "runmqchi .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqlsr .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqxssvn .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "amqztrcn .*-m *X_QUEUE_MANAGER_X( |\$)",
                                        "runmqsc *X_QUEUE_MANAGER_X( |\$)",
                                   ); 
#--------------------------------
# Define WebSphereMQ6 subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String  EntryPointName,
                              IN String  QueueManager,
                              IN Boolean CommandServer,
                              IN String  MQUser,
                              IN String  MQVer,
                              IN String  EnvFile,
                              IN Integer SecondLevelMonitor);

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localQueueManager             WebSphere MQ Queue Manager
 $localCommandServer            WebSphere MQ Command Server
 $localMQUser                   Unix user owner for WebSphere MQ
 $localMQVer                    WebSphere MQ Version
 $localEnvFile                  Environment File
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localQueueManager,
       $localCommandServer,
       $localMQUser,
       $localMQVer,
       $localEnvFile,
       $localSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  QueueManager             [%s]
  CommandServer            [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,  $sEntryPointName,  $localQueueManager, $localCommandServer,   
  $localMQUser, $localMQVer, $localEnvFile, $localSecondLevelMonitor );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;
  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName     QueueManager  CommandServer    MQUser     MQVer
       EnvFile SecondLevelMonitor );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs        = qw ( QueueManager MQUser MQVer );
  my @lsOfflineArgs       = qw ( QueueManager MQUser MQVer );
  my @lsMonitorArgs       = qw ( QueueManager MQUser MQVer );
  my @lsCleanArgs         = qw ( QueueManager MQUser MQVer );
  my @lsImfRegisterArgs   = qw ( QueueManager MQUser MQVer );

  my %hlsNeeded = ( 'online'         => \@lsOnlineArgs,
                    'offline'        => \@lsOfflineArgs,
                    'monitor'        => \@lsMonitorArgs,
                    'clean'          => \@lsCleanArgs,
                    'imf_register'   => \@lsImfRegisterArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #-----------------
  # Validate MQUser
  #-----------------
  my $sUid = getpwnam( $localMQUser );
  if ( ! defined $sUid ) 
  {
      &logError( 20003, "<%s>: The user [%s] is not valid", $sMyName, $localMQUser );
      $iReturnCode = 0 ;
  }
  else {
      &logTrace( 20004, "<%s>: The user [%s] is valid", $sMyName, $localMQUser );
  }

  #-----------------
  # Validate MQVer
  #-----------------
  if ( $localMQVer eq "5.3" || $localMQVer eq "6.0" || $localMQVer eq "7.0")
  {
      &logTrace( 20005, "<%s>: The MQ Version [%s] is valid", $sMyName, $localMQVer );
  }
  else
  {
      &logError( 20006, "<%s>: The MQVersion [%s] is invalid, must be 5.3, 6.0 or 7.0", 
                 $sMyName, $localMQVer );
      $iReturnCode = 0 ;
  }

  $iReturnCode ||
  &logWarn ( 20007,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String QueueManager,
                         IN String MQUser,
                         IN String MQVer,
                         IN String EnvFile );

 Description : Start logic needed for online entry pt

 Parameters :

 $sQueueManager             WebSphere MQ Queue Manager
 $sMQUser                   Unix user owner for WebSphere MQ
 $sMQVer                    WebSphere MQ Version
 $sEnvFile                  Environment File
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20008, "Subroutine <%s> called with:
  QueueManager             [%s]
  CommandServer            [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]",

  $sMyName, $sQueueManager, $bCommandServer, $sMQUser, $sMQVer, $sEnvFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my $sCommand = WebSphereMQ6::START_COMMAND;
  if ( Arch::AIX_OS == Arch::GetOSType() )
  {
    $sCommand = WebSphereMQ6::START_AIX_COMMAND;
  }
  my $sCommandArgs = WebSphereMQ6::START_COMMAND_ARGS;
  $sCommandArgs =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;

  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iRelevantTimeout = 0;
  $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
  if ( $iRelevantTimeout <= 0 )
  {
    &logError( 20009, "<%s>: failed to determine entry point timeout,defaulting to 300", $sMyName);
    $iRelevantTimeout = 300;
  }
  my $iTillWhen = ( $^T + $iRelevantTimeout ) - WebSphereMQ6::EPOP;
  my $iTimeOut  = ($iTillWhen - time);

  my $iExitCode;
  my @lsCommandOut;
  my $bTimedOut;
  my $bSourceEnv   = 0;
  my $sOutFileName = undef;
  my $sFromDir     = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    #----------------------
    # Start program success
    #----------------------
    &logTrace ( 20010, "<%s>: Start command [%s] with args [%s] was successful",
                $sMyName, $sCommand, $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      &logError ( 20011, "<%s>: Start command [%s] with args [%s] timed out",
                  $sMyName, $sCommand, $sCommandArgs);
    }
    else
    {
      &logError ( 20012, "<%s>: Start command [%s] with args [%s] failed to execute",
                  $sMyName, $sCommand, $sCommandArgs);
    }

    #-----------------------------------------------------
    # We need not delay the entry point here. As the
    # online execution was not successful.
    #-----------------------------------------------------
    return 0;
  }

  #-----------------------------------------------------------------
  # Now delay the EP till the resource becomes completely online.
  #-----------------------------------------------------------------
  $iReturnCode = DelayEntryPoint ( $sQueueManager, $bCommandServer, $sMQUser, $sMQVer, $sEnvFile, $iTillWhen );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String QueueManager,
                        IN String MQUser,
                        IN String MQVer,
                        IN String EnvFile );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sQueueManager             WebSphere MQ Queue Manager
 $sMQUser                   Unix user owner for WebSphere MQ
 $sMQVer                    WebSphere MQ Version
 $sEnvFile                  Environment File
 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20013, "Subroutine <%s> called with:
  QueueManager             [%s]
  CommandServer            [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]",

  $sMyName, $sQueueManager, $bCommandServer, $sMQUser, $sMQVer, $sEnvFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-----------------------------------------------------------------
  # We need to first do immediate shutdown.
  # Delay the entry point half way through.
  # If the mq processes are still running, do a preemptive shutdown.
  # Delay the EP till resources goes OFFLINE.
  #-----------------------------------------------------------------

  my $sCommand     = WebSphereMQ6::STOP_COMMAND; 
  my $sCommandArgs = WebSphereMQ6::STOP_IMMEDIATE_COMMAND_ARGS; 
  my $sCommandArgsPreemptive = WebSphereMQ6::STOP_PREEMPTIVE_COMMAND_ARGS;

  if ( Arch::AIX_OS == Arch::GetOSType() )
  {
    $sCommand = WebSphereMQ6::STOP_AIX_COMMAND; 
  }

  $sCommandArgs =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;
  $sCommandArgsPreemptive =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;

  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iRelevantTimeout = 0;
  $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
  if ( $iRelevantTimeout <= 0 )
  {
    &logError( 20014, "<%s>: failed to determine entry point timeout,defaulting to 300", $sMyName);
    $iRelevantTimeout = 300;
  }
  my $iTillWhen = ( $^T + $iRelevantTimeout ) - WebSphereMQ6::EPOP;
  my $iTimeOut  = ($iTillWhen - time)/2;

  my $iExitCode;
  my @lsCommandOut;
  my $bTimedOut;
  my $bSourceEnv   = 0;
  my $sOutFileName = undef;
  my $sFromDir     = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    #--------------------------------------
    # Stop program completed successfully.
    #--------------------------------------
    &logTrace ( 20015, "<%s>: Stop command [%s] with args [%s] executed successfully",
                $sMyName , $sCommand , $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------
      # Stop program timed out.
      #-------------------------
      &logError ( 20016, "<%s>: Stop command [%s] with args [%s] timed out ",
                $sMyName , $sCommand , $sCommandArgs);
    }
    else
    {
      #---------------------------------------------
      # Stop program did not execute successfully.
      #---------------------------------------------
      &logError ( 20017, "<%s>: Stop command [%s] with args [%s] failed to run",
                $sMyName , $sCommand , $sCommandArgs);
      &logWarn ( 20018, "<%s>: Immediate shutdown for [%s] failed, will try Preemptive shutdown.",
                $sMyName , $sQueueManager );
    }

    #---------------------------------------------------------------------
    # Now we should try preemptive shutdown as the immediate shutdown was 
    # timed out or not executed successfully.
    #---------------------------------------------------------------------

    if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgsPreemptive,
                                                 $iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
    {
      #--------------------------------------
      # Stop program completed successfully.
      #--------------------------------------
      &logTrace ( 20019, "<%s>: Stop command [%s] with args [%s] executed successfully",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
    }
    else
    {
      if ( $bTimedOut )
      {
        #-------------------------
        # Stop program timed out.
        #-------------------------
        &logError ( 20020, "<%s>: Stop command [%s] with args [%s] timed out ",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
      }
      else
      {
        #---------------------------------------------
        # Stop program did not execute successfully.
        #---------------------------------------------
        &logError ( 20021, "<%s>: Stop command [%s] with args [%s] failed to run",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
      }

      &logWarn ( 20022, "<%s>: Preemptive shutdown for [%s] failed.",
                $sMyName , $sQueueManager );

      #-----------------------------------------------------
      # We need not delay the entry point here. As the 
      # offline execution was not successful.
      #-----------------------------------------------------
      WebSphereMQ6::DeletePidFile ( );
      return 0;
    }
  }

  #-----------------------------------------------------------------
  # Now delay the EP till the resource becomes completely offline.
  #-----------------------------------------------------------------
  $iReturnCode = DelayEntryPoint ( $sQueueManager, $bCommandServer, $sMQUser, $sMQVer, $sEnvFile, $iTillWhen );

  if ( 0 == $iReturnCode ) {
    my $iCleanReason = VCSDef_CleanReason::OFFLINE_INEFFECTIVE;
    my $bProcCleanup = 0; # Complete clean up..
    if ( WebSphereMQ6::Kill ( $iCleanReason,
                              $sQueueManager,
                              $bCommandServer,
                              $sMQUser,
                              $sMQVer,
                              $sEnvFile,
                              $bProcCleanup ) ) {
      #----------------------
      # Process kill worked..
      #----------------------
      $iReturnCode = 1;
    }
    else {
      #--------------
      # Kill failed..
      #--------------
      &logError ( 20090, "<%s> Failed to kill the processes", $sMyName );
    }
  }
 
  WebSphereMQ6::DeletePidFile ( );
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String QueueManager,
                        IN String MQUser,
                        IN String MQVer,
                        IN String EnvFile
                        IN Boolean ProcCleanup );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason        Clean Reason
 $sQueueManager       WebSphere MQ Queue Manager
 $sMQUser             Unix user owner for WebSphere MQ
 $sMQVer              WebSphere MQ Version
 $sEnvFile            Environment File
 $bProcCleanup        Flag to indicate entire process table scan
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $bProcCleanup ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  $bProcCleanup ||= 0; # Prevent undefs..

  &logTrace ( 20023, "Subroutine <%s> called with:
  CleanReason              [%s]
  QueueManager             [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  ProcCleanup              [%s]",

  $sMyName, $iCleanReason, $sQueueManager, $sMQUser,
  $sMQVer, $sEnvFile, $bProcCleanup );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $iTmpPid, @liPids ) = ();

  if ( VCSDef_MonitorExit::OFFLINE == WebSphereMQ6::GetProcState( $sQueueManager,
                                                                  $bCommandServer,
                                                                  $sMQUser,
                                                                  $sMQVer,
                                                                  $sEnvFile,
                                                                  \@liPids,
                                                                  $bProcCleanup ))
  {
    &logInfo ( 20024, "<%s>: MQ processes already OFFLINE",$sMyName);
  }
  elsif ( VCSDef_CleanReason::OFFLINE_HUNG == $iCleanReason       ||
          VCSDef_CleanReason::OFFLINE_INEFFECTIVE == $iCleanReason )
  {
    #----------------------------------------------
    # Kill the MQ processes in the next block.
    #----------------------------------------------
  }
  else
  {
    my $sCommandArgsPreemptive = WebSphereMQ6::STOP_PREEMPTIVE_COMMAND_ARGS;
    my $sCommand = WebSphereMQ6::STOP_COMMAND; 
    if ( Arch::AIX_OS == Arch::GetOSType() )
    {
      $sCommand = WebSphereMQ6::STOP_AIX_COMMAND; 
    }
    $sCommandArgsPreemptive =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;

    #------------------------------------------
    # Determine the expiration time of the EP.
    #------------------------------------------
    my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
    my $iRelevantTimeout = 0;
    $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
    if ( $iRelevantTimeout <= 0 )
    {
      &logError( 20025, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
      $iRelevantTimeout = 60;
    }
    my $iTillWhen = ( $^T + $iRelevantTimeout ) - WebSphereMQ6::EPOP;
    my $iTimeOut  = $iTillWhen - time;

    my $iExitCode;
    my @lsCommandOut;
    my $bTimedOut;
    my $bSourceEnv   = 0;
    my $sOutFileName = undef;
    my $sFromDir     = undef;

    #------------------------------
    # Try preemptive shutdown now.
    #------------------------------
    &logInfo ( 20026, "<%s>: Initiating preemptive shutdown for [%s]",
               $sMyName, $sQueueManager);

    if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgsPreemptive,
                                                 $iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
    {
      #--------------------------------------
      # Stop program completed successfully.
      #--------------------------------------
      &logTrace ( 20027, "<%s>: Stop command [%s] with args [%s] executed successfully",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
    }
    else
    {
      if ( $bTimedOut )
      {
        #-------------------------
        # Stop program timed out.
        #-------------------------
        &logError ( 20028, "<%s>: Stop command [%s] with args [%s] timed out ",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
      }
      else
      {
        #---------------------------------------------
        # Stop program did not execute successfully.
        #---------------------------------------------
        &logError ( 20029, "<%s>: Stop command [%s] with args [%s] failed to run",
                  $sMyName , $sCommand , $sCommandArgsPreemptive);
      }

      &logWarn ( 20030, "<%s>: Preemptive shutdown for [%s] failed.",
                $sMyName , $sQueueManager );
    }
  }

  #---------------------------
  # Kill MQ Processes..
  #---------------------------
  foreach $iTmpPid (@liPids)
  {
    #-------------------------------
    # Send TERM in the first round.
    #-------------------------------
    if (kill 0, $iTmpPid)
    {
      &logInfo ( 20224, "<%s>: Sending SIGTERM to MQ PID [%s]", 
                 $sMyName, $iTmpPid );
      kill 15, $iTmpPid;
    }
  }
  foreach $iTmpPid (@liPids)
  {
    #----------------------------------------------------------
    # Send KILL if the processes are still found to be running
    #----------------------------------------------------------
    if (kill 0, $iTmpPid )
    {
      &logInfo ( 20225, "<%s>: Sending SIGKILL to MQ PID [%s]", 
                 $sMyName, $iTmpPid );
      kill 9, $iTmpPid;
    }
  }

  #---------------------
  # Clean IPC Resources
  #---------------------
  my $sCommand = WebSphereMQ6::CLEAN_IPC_COMMAND;
  if ( Arch::AIX_OS == Arch::GetOSType() )
  {
    $sCommand = WebSphereMQ6::CLEAN_IPC_AIX_COMMAND;
  }
  
  my $sCommandArgs =  WebSphereMQ6::CLEAN_IPC_COMMAND_ARGS;
  $sCommandArgs    =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;

  my $iExitCode;
  my $bSourceEnv = 0;
  my @lsCommandOut;
  my $bTimedOut,
  my $iTimeOut = 5;
  my $sOutFileName = undef;
  my $sFromDir   = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    #--------------------------------------
    # Stop program completed successfully.
    #--------------------------------------
    &logTrace ( 20208, "<%s>: CleanIPC command [%s] with args [%s] executed successfully",
                $sMyName, $sCommand, $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------
      # Stop program timed out.
      #-------------------------
      &logError ( 20209, "<%s>: CleanIPC command [%s] with args [%s] timed out ",
                $sMyName , $sCommand , $sCommandArgs);
    }
    else
    {
      #---------------------------------------------
      # Stop program did not execute successfully.
      #---------------------------------------------
      &logError ( 20210, "<%s>: CleanIPC command [%s] with args [%s] failed to run",
                $sMyName , $sCommand , $sCommandArgs);
    }

    &logWarn ( 20211, "<%s>: CleanIPC for [%s] failed.",
              $sMyName , $sQueueManager );
  }

  WebSphereMQ6::DeletePidFile ( );
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String QueueManager,
                                   IN String MQUser,
                                   IN String MQVer,
                                   IN String EnvFile,
                                   IN Boolean ProcCleanup );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sQueueManager       WebSphere MQ Queue Manager
 $sMQUser             Unix user owner for WebSphere MQ
 $sMQVer              WebSphere MQ Version
 $sEnvFile            Environment File
 $bProcCleanup        Flag to indicate entire process table scan

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $bProcCleanup ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  $bProcCleanup ||= 0; # Prevent undefs..

  &logTrace ( 20036, "Subroutine <%s> called:
  QueueManager             [%s]
  CommandServer            [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  ProcCleanup              [%s]",

  $sMyName, $sQueueManager, $bCommandServer, $sMQUser,
  $sMQVer, $sEnvFile, $bProcCleanup );

  my ($sLoginName, $sPassword, $iUid, $iGid) = getpwnam($sMQUser);
  if(defined $iUid) 
  {
    #---------------
    # ok - continue
    #---------------
  } 
  else 
  {
    #-----------------------------
    #log error and return unknown
    #-----------------------------

    &logError(20037, "Failed to find UID for user [%s]", $sMQUser);
    return VCSDef_MonitorExit::UNKNOWN;
  }

  my @liPids = ();
  my $iReturnCode = WebSphereMQ6::GetProcState ( $sQueueManager,
                                                 $bCommandServer,
                                                 $sMQUser,
                                                 $sMQVer,
                                                 $sEnvFile,
                                                 \@liPids,
                                                 $bProcCleanup );

  if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20038,
    "<%s> found all required processes",
    $sMyName );
  }
  else {
    &logTrace ( 20039,
    "<%s> did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  &logTrace ( 20040, "<%s> returning [%s]", $sMyName, $iReturnCode );

  return ( $iReturnCode );
}

=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( IN String QueueManager,
                               IN String CommandServer,
                               IN String MQUser,
                               IN String MQVer,
                               OUT List  rliPids,
                               IN Boolean ProcCleanup );

 Description: Performs the first level monitor checks

 Parameters :

 $sQueueManager       WebSphere MQ Queue Manager
 $sCommandServer      WebSphere MQ Command Server
 $sMQUser             Unix user owner for WebSphere MQ
 $sMQVer              WebSphere MQ Version
 $rliOnlinePids       Reference to list of PID hash
 $bProcCleanup        Flag to indicate entire process table scan

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $rliOnlinePids,
       $bProcCleanup ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  $bProcCleanup ||= 0; # Prevent undefs..

  &logTrace ( 20243, "Subroutine <%s> called with:
  QueueManager             [%s]
  CommandServer            [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  rliOnlinePids            [%s]
  ProcCleanup              [%s]",

  $sMyName, $sQueueManager, $bCommandServer, $sMQUser,
  $sMQVer, $sEnvFile, $rliOnlinePids, $bProcCleanup );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid,
       $bRestart, $sLang, $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsKeys, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  $iUid = getpwnam $sMQUser; # Should not fail..
  @$rliOnlinePids = ();      # Reset..


  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( WebSphereMQ6::GetSrvrDetails ( $sMQVer,
                                      $sQueueManager,
                                      $bCommandServer,
                                      $bProcCleanup,
                                      \@lsKeys, # For order..
                                      \%hsSrvrDetails ) ) {

    &logTrace ( 20244,
    "<%s> Got details for monitoring WebSphereMQ version [%s] QueueManager [%s]",
    $sMyName, $sMQVer, $sQueueManager );

    %hsCopy = %hsSrvrDetails;   # Backup..
    $iUseRFLM = scalar @lsKeys; # Set to entire count..
  }
  else {
    &logError ( 20245,
    "<%s> Failed to get details for monitoring WebSphereMQ version [%s] QueueManager [%s]",
    $sMyName, $sMQVer, $sQueueManager );
    return VCSDef_MonitorExit::UNKNOWN;
  }

  #---------------------------------
  # Check if we attempt to be fast..
  #---------------------------------
  unless ( $bProcCleanup ) { # FFLM..
    $iUseRFLM = 0;           # Hopefully skip..

    foreach $sKey ( @lsKeys ) {                    # Maintain order..
      @lsUpdates = ();                             # Reset..
      @lsValues = @{ $hsCopy { $sKey } };          # Get values..
      $sTemp .= "\nWebSphereMQ process [$sKey]\n"; # Display string..

      #----------------
      # Read in pairs..
      #----------------
      while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

        #-----------
        # Run FFLM..
        #-----------
        $iFound = WebSphereMQ6::FFLMStatus ( $sPidFile,
                                             $sProc,
                                             $iUid,
                                             \$iPid );

        #----------------
        # Update status..
        #----------------
        push ( @lsUpdates, $sPidFile, $sProc, $iFound );

        $sTemp .= "PID File     [$sPidFile]\n";
        $sTemp .= "Srch Process [$sProc]\n";
        $sTemp .= "FFLM Status  [$iFound]\n\n";
      }

      @{ $hsCopy { $sKey } } = @lsUpdates;

      #------------------------
      # Handle monitor status..
      #------------------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        push ( @$rliOnlinePids, $iPid ); # Push..
        $hiProcPids { $sKey } = $iPid;   # Store..
      }
      else {
        ++ $iUseRFLM; # Mark it for RFLM..
      }
    }

    #----------------
    # Print summary..
    #----------------
    &logTrace ( 20246,
    "<%s> QueueManager [%s] process details ---%s\n\n",
    $sMyName, $sQueueManager, $sTemp );

    ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
  }
  else { # Use RFLM..
    #-----------------------------------------
    # Mark status of each service as offline..
    #-----------------------------------------
    $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
    foreach $sKey ( keys %hsCopy ) {
      @lsUpdates = (); # Reset..

      @lsValues = @{ $hsCopy { $sKey } }; # Get values..
      while ( scalar @lsValues ) {
        push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
      }

      @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
    }
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20247,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20248,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20249,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20250,
      "<%s> Searching PID [%s] PPID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Owned by MQUser..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $bProcCleanup != 2 ) { # Can be 0/1. Request to filter..
          foreach $sKey ( @lsKeys ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( ( VCSDef_MonitorExit::OFFLINE == $iFound ) &&  # Earlier marked offline..
                   ( $$rhlProc { COMMAND } =~ /$sProc/ ) ) { # Now found..
    
                &logTrace ( 20251,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );
                $hiProcPids { $sKey } = $iPid;        # Store..
                unshift ( @$rliOnlinePids, $iPid );   # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20252,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20253,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( WebSphereMQ6::CreatePidFile ( $sPidFile,
                                           $hiProcPids { $sKey } ) ) {
          &logTrace ( 20254,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20255,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        if ( $sPidFile =~ /pidfile.amqpcsea$/         && # Command server down.. 
             ( "6.0" eq $sMQVer || "7.0" eq $sMQVer ) && # For the valid version..
             $bCommandServer ) {                         # Request startup..
          ++ $bRestart; # Mark for restart..
        }
        else {
          $iOffline = $iFound;
        }
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN; # Verify..

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20256,
    "<%s> Found [%s] PIDs [%s] for QueueManager [%s].
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $sQueueManager, $iReturnCode );
  }

  if ( ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) &&
         $bRestart ) {
    &logWarn ( 20257,
    "<%s> Did not find command server process running. Requesting start up..",
    $sMyName );

    if ( WebSphereMQ6::StartCmdServer ( $sQueueManager, # Need to check the signature..
                                        $sEnvFile,
                                        $sMQUser ) ) {
      &logInfo ( 20258,
      "<%s> Command server started successfully.", $sMyName );
    }
    else {
      &logWarn ( 20259,
      "<%s> Command server could not be started.", $sMyName );
    }
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  &logTrace ( 20260,
  "<%s> WebSphereMQ QueueManager [%s] process details ---\n%s\nReturn code is [%s]",
  $sMyName, $sQueueManager, $sTemp, $iReturnCode );
  $sTemp = undef; # Release..

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String QueueManager,
                          IN String MQUser,
                          IN String MQVer,
                          IN String EnvFile );

 Description: Performs the in-depth monitoring

 Parameters :

 $sQueueManager             WebSphere MQ Queue Manager
 $sMQUser                   Unix user owner for WebSphere MQ
 $sMQVer                    WebSphere MQ Version
 $sEnvFile                  Environment File
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $iSecondLevelMonitor,
       $sMonitorProgram,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20046, "Subroutine <%s> called with:
  QueueManager             [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]
  level_two                [%s]",
  $sMyName, $sQueueManager, $sMQUser, $sMQVer, $sEnvFile, $iSecondLevelMonitor, $sMonitorProgram, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iRelevantTimeout = 0;
  my $iTillWhen = ();

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( -300 == $level_two ) {
     $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );  
     &logTrace ( 20305, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }

  if ( $level_two >= 1 ) 
  {

    $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
    if ( $iRelevantTimeout <= 0 )
    {
      &logError( 20047, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
      $iRelevantTimeout = 60;
    }
    $iTillWhen = ( $^T + $iRelevantTimeout ) - WebSphereMQ6::EPOP;

    &logTrace ( 20048,
    "<%s> Proceeding with second level monitoring", $sMyName );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         WebSphereMQ6::SecondLevelMonitor ( $sQueueManager,
                                            $sMQUser,
                                            $sMQVer,
                                            $sEnvFile,
                                            $iTillWhen ) ) ) 
   {
      #--------------------------------------------
      # LOG TRACE MESSAGE THAT INSTANCE IS ONLINE..
      #--------------------------------------------
      &logTrace ( 20309,
      "<%s> WebSphereMQ instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      #------------------------------------------------
      # LOG ERROR MESSAGE THAT INSTANCE IS NOT ONLINE..
      #------------------------------------------------
      &logError ( 20310,
      "<%s> WebSphereMQ instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  if ( $sMonitorProgram )
  {
    $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
    if ( $iRelevantTimeout <= 0 )
    {
      &logError( 20217, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
      $iRelevantTimeout = 60;
    }
    $iTillWhen = ( $^T + $iRelevantTimeout ) - WebSphereMQ6::EPOP;
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) 
  {
    $iReturnCode = WebSphereMQ6::ExternalMonitor( $sExtCmd,
                                                  $sMQUser,
                                                  $sExtCmdArgs,
                                                  $sEnvFile,
                                                  $iTillWhen);
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String QueueManager,
                                     IN String MQUser,
                                     IN String MQVer,
                                     IN String EnvFile,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sQueueManager             WebSphere MQ Queue Manager
 $sMQUser                   Unix user owner for WebSphere MQ
 $sMQVer                    WebSphere MQ Version
 $sEnvFile                  Environment File
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sQueueManager,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $iTillWhen ) = @_ ;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20049, "Subroutine <%s> called with:
  QueueManager             [%s]
  MQUser                   [%s]
  MQVer                    [%s]
  EnvFile                  [%s]
  TillWhen                 [%s]",

  $sMyName,   $sQueueManager,   $sMQUser,    $sMQVer,    $sEnvFile,
  $iTillWhen );  


  my $iTimeOut  = $iTillWhen - time;

  my $sCommand = WebSphereMQ6::SECOND_LEVEL_MONITOR_COMMAND;
  if ( Arch::AIX_OS == Arch::GetOSType() )
  {
    $sCommand = WebSphereMQ6::SECOND_LEVEL_MONITOR_AIX_COMMAND;
  }
  my $sCommandArgs = WebSphereMQ6::SECOND_LEVEL_MONITOR_COMMAND_ARGS;
  $sCommandArgs =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;

  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  my $bSourceEnv = 0;
  my $iExitCode;
  my @lsCommandOut = ();
  my $bTimedOut;
  my $sOutFileName = undef;
  my $sFromDir = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 \$iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    if ( 0 == $iExitCode )
    {
      #------------------------
      # SLM ran successfully.
      #------------------------
      &logTrace ( 20050, "<%s>: SecondLevelMonitor executed successfully",$sMyName);
    }
    else
    {
      #------------------------
      # SLM failed..
      #------------------------
      &logError ( 20051, "<%s>: SecondLevelMonitor failed with return code [%s]",
                  $sMyName, $iExitCode);
      return VCSDef_MonitorExit::OFFLINE
    }
  }
  else
  {
    if ( $bTimedOut )
    {
      #------------------------
      # SLM command timed out.
      #------------------------
      &logError ( 20052, "<%s>: SecondLevelMonitor timed out.",$sMyName);
      return VCSDef_MonitorExit::TIMEOUT;
    }
    else
    {
      #------------------------
      # SLM command failed.
      #------------------------
      &logError ( 20053, "<%s>: SecondLevelMonitor command failed",$sMyName);
    }
    return VCSDef_MonitorExit::UNKNOWN;
  }
  
  return VCSDef_MonitorExit::ONLINE;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String ExtCmd,
                                  IN String ExtCmdArgs,
                                  IN String EnvFile,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sExtCmd                   User defined utilty to execute
 $sExtCmdArgs               Command line arguments
 $sEnvFile                  Environment file to be sourced
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sExtCmd,
       $sMQUser,
       $sExtCmdArgs,
       $sEnvFile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20054, "Subroutine <%s> called with:
  ExtCmd                   [%s]
  ExtCmdArgs               [%s]
  EnvFile                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sExtCmd, $sExtCmdArgs, $sEnvFile, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeOut, @lsCmdOut ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sExtCmd ) {
   &logError ( 20055,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sExtCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sExtCmd ) {
   &logError ( 20056,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sExtCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeOut = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20057,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sExtCmd, abs ( $iTimeOut ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20058,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sExtCmd, $sExtCmdArgs, $iTimeOut );

  my $bSourceEnv = 0;
  my @lsCommandOut = ();
  my $bTimedOut;
  my $sOutFileName = undef;
  my $sFromDir = undef;

  #-------------------------------------------------
  # EXECUTE THE COMMAND USING ACCLIB FUNCTIONS..
  #-------------------------------------------------

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sExtCmd,
                                                 $sExtCmdArgs,
                                                 \$iExitCode,
                                                 $sMQUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    if ( 0 == $iExitCode || VCSDef_MonitorExit::ONLINE == $iExitCode )
    {
      #------------------------------------------------------------
      # External MonitorProgram reported MQ resource to be ONLINE
      #------------------------------------------------------------
      &logTrace ( 20059, "<%s>: MonitorProgram [%s] with arguments [%s] reporting MQ ONLINE, exit code [%s]",
                  $sMyName , $sExtCmd , $sExtCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::ONLINE;
    }
    elsif ( 1 == $iExitCode || VCSDef_MonitorExit::OFFLINE == $iExitCode )
    {
      #------------------------------------------------------------
      # External MonitorProgram reported MQ resource to be OFFLINE
      #------------------------------------------------------------
      &logError ( 20060, "<%s>: MonitorProgram [%s] with arguments [%s] reporting MQ OFFLINE, exit code [%s]",
                  $sMyName , $sExtCmd , $sExtCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    else
    {
      #------------------------------------------------------------
      # External MonitorProgram reported MQ resource to be UNKNOWN
      #------------------------------------------------------------
      &logError ( 20061, "<%s>: MonitorProgram [%s] with arguments [%s] reporting MQ UNKNOWN, exit code [%s]",
                  $sMyName , $sExtCmd , $sExtCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }
  else
  {
    if ( $bTimedOut )
    {
      #----------------------------------------------------------------------
      # Command timed out!  Return UNKNOWN as the first (and possibly second)
      # level monitoring did work -- it is considered safer to return UNKNOWN
      # as the service may be generally OK, and the MonitorProgram didn't
      # determine for sure that the service was not up, as the defined
      # methodology to do so (as implemented in the MonitorProgram) could not
      # be run successfully (it timed out).
      #----------------------------------------------------------------------
      &logError ( 20062, "<%s>: MonitorProgram [%s] with arguments [%s] timed out",
                  $sMyName , $sExtCmd , $sExtCmdArgs );
      $iReturnCode = VCSDef_MonitorExit::TIMEOUT;
    }
    else
    {
      &logError ( 20063, "<%s>: MonitorProgram [%s] with arguments [%s] did not execute successfully, exit code[%s]",
                  $sMyName , $sExtCmd , $sExtCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }

  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int GetEPTimeout ( IN String EntryPointName
                               IN String ResourceType  );

 Description : Returns the Entry Point Timeout for the EP which we are in

 Parameters :

 $sEntryPointName           The EP which we are currently running
 $sResourceType             VCS resource type

 Returns  : Integer value as defined below:

 -------------------------------------------------------------------------
=cut

sub GetEPTimeout
{
  my ( $sEntryPointName, $sResourceType) = @_;

  #------------------
  # Announce Entry..
  #------------------

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20064, "Subroutine <%s> called with:
  EntryPointName [%s]
  ResourceType   [%s]", 
  $sMyName , $sEntryPointName , $sResourceType );

  my $sObjectName          = "hatype";
  my %hTypeValues;

  my $iEngineTimeout;
  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20089,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
    return ( $iEngineTimeout ); 
  }

  if ( ! VCSagentFW::GetVCSObjectHash($sObjectName,$sResourceType,\%hTypeValues) )
  {
    &logError(20065,"<%s>:GetVCSObjectHash call failed.",$sMyName);
    return 0;
  }

  #-------------------------------
  # Make "monitor" to "Monitor"
  # offline to Offline and so on
  #-------------------------------

  my $sTempEntry =  $sEntryPointName ;
  $sTempEntry    =~ s/(\w)(.*)$/\U$1\L$2/; 

  &logTrace(20066,"<%s>: Returning [%s] value [%s]",$sMyName, $sTempEntry."Timeout",$hTypeValues{$sTempEntry."Timeout"});
  return ( $hTypeValues{$sTempEntry."Timeout"});
}

=pod

 -------------------------------------------------------------------------
 Function : int DelayEntryPoint ( IN String QueueManager,
                                  IN String MQUser,
                                  IN String MQVer,
                                  IN String EnvFile
                                  IN Integer DelayToTime)

 Description : Returns the Entry Point Timeout for the EP which we are in.

 If we are in online , the subroutine delays the entry point till all 
 required MQ processes start running.

 If we are in offline , the subroutine delays the entry point till all 
 required MQ processes stop running.

 Parameters :

 $sQueueManager             WebSphere MQ Queue Manager
 $sMQUser                   WebSphere MQ User
 $sMQVer                    WebSphere MQ Version
 $sEnvFile                  Environment File
 $iDelayToTime              Max time allowed to delay the EP

 Returns  : Integer value as defined below:
 0 Failure
 1 Success

 -------------------------------------------------------------------------

=cut

sub DelayEntryPoint
{
  my ( $sQueueManager,
       $bCommandServer,
       $sMQUser,
       $sMQVer,
       $sEnvFile,
       $iDelayToTime )   = @_ ;

  #------------------
  # Announce Entry..
  #------------------

  my $sMyName = ( caller 0)[3];
  &logTrace ( 20067, "Subroutine <%s> called with:
  QueueManager   [%s]
  MQUser         [%s]
  MQVer          [%s]
  DelayToTime    [%s]",
  $sMyName, $sQueueManager, $sMQUser, $sMQVer, $iDelayToTime );

  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  my ( $bProcCleanup, $iReturnCode, @liPids ) = ();
  my $iNSR = $iDelayToTime - time;

  while ( $iNSR > CHECK_INTERVAL )
  {
    if ( $sEntryPointName eq "offline" )
    {
      #-------------------------------------------------------------
      # We are in offline. Check whether MQ processes are running.
      # If MQ processes are running, sleep for 5 seconds.
      # Repeat above two steps till MQ processes are gone.
      #-------------------------------------------------------------
      $bProcCleanup = 1; # Full scan..
      $iReturnCode = WebSphereMQ6::GetProcState ( $sQueueManager,
                                                  $bCommandServer,
                                                  $sMQUser,
                                                  $sMQVer,
                                                  $sEnvFile,
                                                  \@liPids,
                                                  $bProcCleanup );
      if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode )
      {
        &logTrace ( 20069, "<%s>: MQ processes not running", $sMyName);
        return 1;
      }
      else
      {
        &logInfo ( 20070, "<%s>: MQ processes still running. Sleeping for 5 seconds", $sMyName);
        sleep CHECK_INTERVAL;
        $iNSR = $iDelayToTime - time;
      }
    } 
    elsif ( $sEntryPointName eq "online" )
    {
      #-------------------------------------------------------------
      # We are in online. Check whether MQ processes are running.
      # If all MQ processes are NOT running, sleep for 5 seconds.
      # Repeat above two steps till all MQ processes are ONLINE.
      #-------------------------------------------------------------
      $bProcCleanup = 0; # Be fast..
      $iReturnCode = WebSphereMQ6::GetProcState ( $sQueueManager,
                                                  $bCommandServer,
                                                  $sMQUser,
                                                  $sMQVer,
                                                  $sEnvFile,
                                                  \@liPids,
                                                  $bProcCleanup );
      if ( VCSDef_MonitorExit::ONLINE == $iReturnCode )
      {
        &logTrace ( 20071, "<%s>: All MQ processes are ONLINE", $sMyName);
        return 1;
      }
      else
      {
        &logInfo ( 20072, "<%s>: All MQ processes are not ONLINE. Sleeping for 5 seconds.", $sMyName);
        sleep CHECK_INTERVAL;
        $iNSR = $iDelayToTime - time;
      }
    }
  }

  if ( $iNSR <= 5 )
  {
    &logInfo ( 20068, "<%s>: Not enough time left to do MQ process checks", $sMyName);
  }

  return 0;
}


=pod

 -------------------------------------------------------------------------
 Function : int HandlePartialOnline ( IN Integer ResourceState )

 Description : Returns Exit Code to be returned in case of the Resource
 found to be Partial Online.

 Parameters :

 $iResourceState            ResourceState as identified by the Monitor

 Returns  : Integer value as defined below:
 VCSDef_MonitorExit::OFFLINE
 VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub HandlePartialOnline
{
  my ( $iResourceState  ) = @_;
  
  #------------------
  # Announce Entry..
  #------------------

  my $sMyName = ( caller 0)[3];
  &logTrace ( 20242, "Subroutine <%s> called with:
  ResourceState   [%s]",
  $sMyName, $iResourceState);

  my $sResourceName   = $ENV{VCS_LOG_RESOURCE_NAME};

  #-----------------------------------------------------------------------------
  # First Case:
  # Current resource State is (a kind of) OFFLINE, but we were trying to
  # go online (in one of it's various forms), and yet we only found some
  # resource components running.  This means the attempt to ONLINE failed,
  # and therefore we should provide the VCS engine an exit code of OFFLINE
  # to indicate it failed.
  #-----------------------------------------------------------------------------

  if( ( VCSDef_ResourceState::OFFLINE               == $::giState ||
        VCSDef_ResourceState::OFFLINE_STATE_UNKNOWN == $::giState ||
        VCSDef_ResourceState::FAULTED_RESTARTING    == $::giState
      ) && (
        VCSDef_ResourceIState::WAITING_TO_GO_ONLINE                   == $::giIState ||
        VCSDef_ResourceIState::WAITING_TO_GO_ONLINE_REVERSE           == $::giIState ||
        VCSDef_ResourceIState::WAITING_TO_GO_ONLINE_REVERSE_PROPAGATE == $::giIState ||
        VCSDef_ResourceIState::WAITING_TO_GO_ONLINE_REVERSE_PATH      == $::giIState ))
  {
    return VCSDef_MonitorExit::OFFLINE;
  }

  #-----------------------------------------------------------------------------
  # Second Case:
  # Current resource State is (a kind of) ONLINE, and we were not
  # trying to go offline, and yet we only found some resource components
  # running.  This effectively means that the resource has partially
  # faulted. However, as there is no a way to indicate a partial fault,
  # we have to fault the resource instance, by returning OFFLINE, which
  # will cause it to call the clean entry point with clean reason:
  # UNEXPECTED_OFFLINE.  If configured appropriately, it will then get
  # restarted via the normal VCS mechanisms.
  #-----------------------------------------------------------------------------
  if( ( VCSDef_ResourceState::ONLINE                   == $::giState ||
        VCSDef_ResourceState::ONLINE_STATE_UNKNOWN     == $::giState ||
        VCSDef_ResourceState::ONLINE_MONITOR_TIMEDOUT  == $::giState ||
        VCSDef_ResourceState::ONLINE_UNABLE_TO_OFFLINE == $::giState
      ) && (
        VCSDef_ResourceIState::NOT_WAITING == $::giIState ) )
  {
    return VCSDef_MonitorExit::OFFLINE;
  }
 
    #-----------------------------------------------------------------------------
  # Third Case:
  # Current resource State is (a kind of) OFFLINE, and we were not trying to
  # go online (in one of it's various forms), and yet we only found some
  # resource components running.  This means that we are partially online and
  # we need to clean up the environment so that we can attempt to completely
  # online at some future time.  Also, we need to make sure that any possibly
  # dependent resources do not attempt to online while we are only in a
  # partially online state!
  #-----------------------------------------------------------------------------
  if( ( VCSDef_ResourceState::OFFLINE               == $::giState ||
        VCSDef_ResourceState::OFFLINE_STATE_UNKNOWN == $::giState ||
        VCSDef_ResourceState::FAULTED_RESTARTING    == $::giState
      ) && (
        VCSDef_ResourceIState::NOT_WAITING == $::giIState ) )
  {
    #-----------------------------------------------------------------------------
    # Queue up an "hares -probe" on this resource.  This will most likely find
    # the resource PARTIAL_ONLINE again, but this next time it will return OFFLINE
    # due to the algorithms implemented above (Refer to "Second Case"
    # explanation above).
    #-----------------------------------------------------------------------------
    my $sMyVCS_SysName = "UNKNOWN";
    if( Sys::GetSystemNameAsVcsKnowsIt( \$sMyVCS_SysName ) )
    {
      my $sEnvFile         = "";   # no EnvFile required
      my $bSourceUserEnv   = 0;    # Do NOT source the user's login profile file
      my $sCommand         = "";
      if ( defined $ENV{"CLUSTER_HOME"} )
      {
        $sCommand         = $ENV{'CLUSTER_HOME'} . "/bin/hares";
      }
      else { $sCommand      = $ENV{'VCS_HOME'} . "/bin/hares";}
      my $sCommandArgs     = "-probe $sResourceName -sys $sMyVCS_SysName";
      my $iCommandExitCode = 1;    # set by called sub
      my $sUser            = "root";  # must be root to run ha commands
      my @lsCommandOutput  = ();     # set by called sub
      my $bTimedOut        = 1;    # set by called sub
      my $iTimeout         = 10;   # SWAG at a reasonable maximum time to run the VCS command
      my $sRunFromDir      = "";   # do not chdir in order to run from a specified directory

      if( Sys::RunWithEnvCmdWithOutputWithTimeOut( $sEnvFile,
                                                   $bSourceUserEnv,
                                                   $sCommand,
                                                   $sCommandArgs,
                                                   \$iCommandExitCode,
                                                   $sUser,
                                                   \@lsCommandOutput,
                                                   \$bTimedOut,
                                                   $iTimeout,
                                                   $sRunFromDir      ) )
      {
        if( 0 == $iCommandExitCode )
        {
          &logInfo( 20203, "Probed resource [%s], putting it back into the probe queue.  If the resource is again found to be PARTIAL_ONLINE, it will begin FAULT processing",$sResourceName );
        }
        else
        {
          &logError( 20204, "Problems probing resource [%s]", $sResourceName );
          VCSagentFW::logOutputFromCommand( "$sCommand $sCommandArgs",
                                                    $iCommandExitCode,
                                                    \@lsCommandOutput );
        }
      }
      else
      {
        #-----------------------------------------------
        # first check to see if the command timed out...
        #-----------------------------------------------
        if( $bTimedOut )
        {
          &logError( 20205, "The command line [%s] did not complete within the allotted amount of time ([%s] seconds)",
                              "$sCommand $sCommandArgs",
                              $iTimeout );
        }
        else
        {
          #---------------------------------------------------------------------
          # log error & return 0 to indicate internal failure in attempted call
          # to RunWithEnvCmdWithOutputWithTimeOut...
          #---------------------------------------------------------------------
          &logError( 20206, "Errors encountered during call to RunWithEnvCmdWithOutputWithTimeOut");
        }
      }
    }
    else
    {
      &logError( 20207, "Could not get VCS SysName");
    }
    #----------------------
    # Generate trace dump.
    #----------------------

    return VCSDef_MonitorExit::PARTIAL_ONLINE;
  }
  
  return VCSDef_MonitorExit::PARTIAL_ONLINE;

}

=pod

 -------------------------------------------------------------------------
 Function : void StartCmdServer ( IN String QueueManager
                                  IN String EnvFile
                                  IN String MQUser );

 Description : Checks whether command server process is running. Restart
 the command server if found to be offline.

 Parameters:

 $sQueueManager             WebSphere MQ Queue Manager
 $sEnvFile                  Environment File
 $sMQUser                   WebSphere MQ User


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub StartCmdServer {
  my ( $sQueueManager,
       $sEnvFile,
       $sMQUser ) = @_;

  #------------------
  # Announce Entry..
  #------------------
  my $sMyName = ( caller 0)[3];
  &logTrace ( 20261, "Subroutine <%s> called with:
  QueueManager   [%s]
  EnvFile        [%s]
  MQUser         [%s]",

  $sMyName, $sQueueManager, $sEnvFile, $sMQUser );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCommand = $^O eq 'aix' ? WebSphereMQ6::START_COMMAND_AIX_CMDSERVER :
                                WebSphereMQ6::START_COMMAND_CMDSERVER;
  my $sCommandArgs =  $sQueueManager;
  my $bSourceEnv = 0;
  my $bTimedOut = 0;
  my $iTimeOut = 5;
  my ( $riExitCode, @lsCommandOut ) = ();
  my $iReturnCode = 1; # Hopeful success..

  if ( $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                                $bSourceEnv,
                                                                $sCommand,
                                                                $sCommandArgs,
                                                                \$riExitCode,
                                                                $sMQUser,
                                                                \@lsCommandOut,
                                                                \$bTimedOut,
                                                                $iTimeOut ) ) {
    #--------------------------------------
    # Start program completed successfully.
    #--------------------------------------
    if ( $riExitCode ) { # Failure..
      &logError ( 20262,
      "<%s> Command [%s] with args [%s] returned exit code [%s].",
      $sMyName, $sCommand, $sCommandArgs, $riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20263,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $riExitCode );
    }
  }
  else {
    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20264,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeOut );
    }
    else { # Failure due to something else..
      &logError ( 20265,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }

    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20266, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20267, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20268,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20269,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20270,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
  WebSphereMQ6::DeletePidFile ( );
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

  &logTrace ( 20311, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20312,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "/tmp/.VRTSWebSphereMQ6/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20313,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20314,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20315,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20271, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20272, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20273, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20274, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20275,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20276,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20277,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( WebSphereMQ6::GetPidDetails ( $iPid,
                                       \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20278,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20279,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20280,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20281,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20282, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20283, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20284,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20285, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20286,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20287,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String MQVer,
                                  IN String QueueManager,
                                  IN Boolean CommandServer,
                                  IN Boolean CommandServer,
                                  OUT RefList SrvrKeys,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sMQVer               WebSphere MQ Version
 $sQueueManager        WebSphere MQ Queue Manager
 $bCommandServer       Flag to indicate monitor the WebSphere MQ Command Server
 $bProcCleanup         Flag to indicate entire process table scan
 $rlSrvrKeys           Stores keys for accessing rhlSrvrDetails in correct order
 $rhlSrvrDetails       Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sMQVer,
       $sQueueManager,
       $bCommandServer,
       $bProcCleanup,
       $rlSrvrKeys,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20288, "Subroutine <%s> called with:
  MQVer                   [%s]
  QueueManager            [%s]
  CommandServer           [%s]
  ProcCleanup             [%s]
  SrvrKeys                [%s]
  SrvrDetails             [%s]",

  $sMyName, $sMQVer, $sQueueManager, $bCommandServer,
  $bProcCleanup, $rlSrvrKeys, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $sEntryPointName, @lsSrchPatterns,
       %hsDup ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #--------------------------
  # List of search patterns..
  #--------------------------
  if ( "clean" eq $sEntryPointName   ||
       "offline" eq $sEntryPointName || $bProcCleanup ) {
    if ( $sMQVer eq "5.3" ) {
      @lsSrchPatterns = WebSphereMQ6::KILL_PROC_LIST_53;
    }
    elsif ( $sMQVer eq "7.0" ) {
      @lsSrchPatterns = WebSphereMQ6::KILL_PROC_LIST_70;
    }
    else {
      @lsSrchPatterns = WebSphereMQ6::KILL_PROC_LIST_60;
    }
  }
  else {
    if ( $sMQVer eq "5.3" ) {
      @lsSrchPatterns  = WebSphereMQ6::MONITOR_PROC_LIST_53;
    }
    elsif ( $sMQVer eq "7.0" ) {
      @lsSrchPatterns  = WebSphereMQ6::MONITOR_PROC_LIST_70;
    }
    else {
      @lsSrchPatterns  = WebSphereMQ6::MONITOR_PROC_LIST_60;
    }
  }

  #---------------------------------------
  # Monitor the command server, if asked..
  #---------------------------------------
  %hsDup = map { m/(\S+)/; $1 => 1 } @lsSrchPatterns; # Create unique..
  WebSphereMQ6::COMMAND_SERVER_PROC =~ m/(\S+)/;

  if ( $bCommandServer && ! $hsDup { $1 } ) {
    &logTrace ( 20289,
    "<%s> Command server requested to be monitored via CommandServer [%s].",
    $sMyName, $bCommandServer );
    push @lsSrchPatterns, WebSphereMQ6::COMMAND_SERVER_PROC;
  }

  %hsDup = (); # Free..

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20290, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20291, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  ( @$rlSrvrKeys, %$rhlSrvrDetails ) = (); # Truncate..
  foreach $sProc ( @lsSrchPatterns ) {
    $sProc =~ s/X_QUEUE_MANAGER_X/$sQueueManager/;
    ( $sKey ) = $sProc =~ m/(\S+)/;

    #-----------------------------
    # Skip listener if not asked..
    #-----------------------------
    if ( $sKey eq 'runmqlsr' ) {
      if ( ! $::giMonitorListener ) {
        &logTrace ( 20292,
        "<%s> Skipping monitoring of listener process", $sMyName );
        next;
      }
      else {
        &logTrace ( 20293,
        "<%s> Listener process requested to be monitored via MonitorListener [%s]",
        $sMyName, $::giMonitorListener );
      }
    }

    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
    push @$rlSrvrKeys, $sKey; # Store key..
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
